@@ -1,9 +1,9 @@
 """
-uttp - pretty simple and powerful web server for tiny platforms like ESP8266 / ESP32
+Tiny Web - pretty simple and powerful web server for tiny platforms like ESP8266 / ESP32
 MIT license
-(C) Derek Anderson 2022
 (C) Konstantin Belyalov 2017-2018
 """
+import logging
 import uasyncio as asyncio
 import uasyncio.core
 import ujson as json
@@ -14,7 +14,7 @@ import uerrno as errno
 import usocket as socket
 
 
-__version__ = (0,1,0)
+log = logging.getLogger('WEB')
 
 type_gen = type((lambda: (yield))())
 
@@ -114,6 +114,7 @@ class request:
             if line == b'\r\n':
                 break
             frags = line.split(b':', 1)
+            frags[0] = frags[0].lower()
             if len(frags) != 2:
                 raise HTTPException(400)
             if frags[0] in save_headers:
@@ -132,10 +133,10 @@ class request:
         # chunks instead of accumulating payload.
         gc.collect()
         # Parse payload depending on content type
-        if b'Content-Type' not in self.headers:
-            # unknown Content-Type
-            raise HTTPException(400)
-        size = int(self.headers[b'Content-Length']) if b'Content-Length' in self.headers else None
+        if b'content-type' not in self.headers:
+            # Unknown content type, return unparsed, raw data
+            return {}
+        size = int(self.headers[b'content-length']) if b'content-length' in self.headers else None
         if size is not None:
             if size > self.params['max_body_size'] or size < 0:
                 raise HTTPException(413)
@@ -149,7 +150,7 @@ class request:
                 raise HTTPException(413)
         # Use only string before ';', e.g:
         # application/x-www-form-urlencoded; charset=UTF-8
-        ct = self.headers[b'Content-Type'].split(b';', 1)[0]
+        ct = self.headers[b'content-type'].split(b';', 1)[0]
         try:
             if ct == b'application/json':
                 return json.loads(data)
@@ -158,6 +159,8 @@ class request:
         except ValueError:
             # Re-generate exception for malformed form data
             raise HTTPException(400)
+    
+    json = read_parse_form_data
 
 
 class response:
@@ -208,6 +211,16 @@ class response:
         await self._send_headers()
         if msg:
             await self.send(msg)
+    
+    async def _handle_return(self, o):
+        if o is None:
+          return
+        elif isinstance(o, dict):
+          self.add_header('Content-Type', 'application/json')
+          await self._send_headers()
+          await self.send(json.dumps(o))
+        else:
+          raise Exception('unknown return type: %s' % o)
 
     async def redirect(self, location, msg=None):
         """Generate HTTP redirect response to 'location'.
@@ -397,6 +410,7 @@ class webserver:
         self.explicit_url_map = {}
         self.catch_all_handler = None
         self.parameterized_url_map = {}
+        self.regex_urls = []
         # Currently opened connections
         self.conns = {}
         # Statistics
@@ -416,6 +430,15 @@ class webserver:
             # Save parameter into request
             req._param = req.path[idx:].decode()
             return self.parameterized_url_map[path2]
+        
+        # Third try - see if it matches any regex urls
+        for pattern, f, params in self.regex_urls:
+            match = pattern.match(req.path)
+            if match:
+              print('match', match)
+              num_groups = int(repr(match).split('=')[-1].rstrip('>'))
+              req._params = [match.group(i).decode() for i in range(1, num_groups)]
+              return f, params
 
         if self.catch_all_handler:
             return self.catch_all_handler
@@ -427,6 +450,7 @@ class webserver:
         await req.read_request_line()
         # Find URL handler
         req.handler, req.params = self._find_url_handler(req)
+        print(req.method.decode(), req.path.decode())
         if not req.handler:
             # No URL handler found - read response and issue HTTP 404
             await req.read_headers()
@@ -468,9 +492,12 @@ class webserver:
             # Handle URL
             gc.collect()
             if hasattr(req, '_param'):
-                await req.handler(req, resp, req._param)
+                ret = await req.handler(req, resp, req._param)
+            elif hasattr(req, '_params'):
+                ret = await req.handler(*[req, resp]+req._params)
             else:
-                await req.handler(req, resp)
+                ret = await req.handler(req, resp)
+            await resp._handle_return(ret)
             # Done here
         except (asyncio.CancelledError, asyncio.TimeoutError):
             pass
@@ -481,15 +508,18 @@ class webserver:
                 try:
                     await resp.error(500)
                 except Exception as e:
+                    log.exc(e, "")
                     sys.print_exception(e)
         except HTTPException as e:
             try:
                 await resp.error(e.code)
             except Exception as e:
+                log.exc(e)
                 sys.print_exception(e)
         except Exception as e:
             # Unhandled expection in user's method
-            print(req.path.decode())
+            log.error(req.path.decode())
+            log.exc(e, "")
             sys.print_exception(e)
             try:
                 await resp.error(500)
@@ -521,8 +551,7 @@ class webserver:
             allowed_access_control_headers - Default value for the same name header. Defaults to *
             allowed_access_control_origins - Default value for the same name header. Defaults to *
         """
-        if url == '' or '?' in url:
-            raise ValueError('Invalid URL')
+            
         # Initial params for route
         params = {'methods': ['GET'],
                   'save_headers': ['Content-Length', 'Content-Type'],
@@ -534,21 +563,29 @@ class webserver:
         params['allowed_access_control_methods'] = ', '.join(params['methods'])
         # Convert methods/headers to bytestring
         params['methods'] = [x.encode() for x in params['methods']]
-        params['save_headers'] = [x.encode() for x in params['save_headers']]
-        # If URL has a parameter
-        if url.endswith('>'):
-            idx = url.rfind('<')
-            path = url[:idx]
-            idx += 1
-            param = url[idx:-1]
-            if path.encode() in self.parameterized_url_map:
-                raise ValueError('URL exists')
-            params['_param_name'] = param
-            self.parameterized_url_map[path.encode()] = (f, params)
+        params['save_headers'] = [x.lower().encode() for x in params['save_headers']]
 
-        if url.encode() in self.explicit_url_map:
-            raise ValueError('URL exists')
-        self.explicit_url_map[url.encode()] = (f, params)
+        if url.__class__.__name__=='ure':
+          # url is a regex
+            self.regex_urls.append((url, f, params))
+        else:
+          if url == '' or '?' in url:
+              raise ValueError('Invalid URL')
+
+          # If URL has a parameter
+          if url.endswith('>'):
+              idx = url.rfind('<')
+              path = url[:idx]
+              idx += 1
+              param = url[idx:-1]
+              if path.encode() in self.parameterized_url_map:
+                  raise ValueError('URL exists')
+              params['_param_name'] = param
+              self.parameterized_url_map[path.encode()] = (f, params)
+
+          if url.encode() in self.explicit_url_map:
+              raise ValueError('URL exists')
+          self.explicit_url_map[url.encode()] = (f, params)
 
     def add_resource(self, cls, url, **kwargs):
         """Map resource (RestAPI) to URL
@@ -614,6 +651,14 @@ class webserver:
             self.add_route(url, f, **kwargs)
             return f
         return _route
+    
+    def get(self, url, **kwargs):
+        kwargs['methods'] = ['GET']
+        return self.route(url, **kwargs)
+
+    def post(self, url, **kwargs):
+        kwargs['methods'] = ['POST']
+        return self.route(url, **kwargs)
 
     def resource(self, url, method='GET', **kwargs):
         """Decorator for add_resource() method
@@ -695,3 +740,4 @@ class webserver:
         asyncio.cancel(self._server_coro)
         for hid, coro in self.conns.items():
             asyncio.cancel(coro)
+
